@@ -438,18 +438,6 @@ select arrayrange(ARRAY[2,1], ARRAY[1,2]);  -- fail
 select array[1,1] <@ arrayrange(array[1,2], array[2,1]);
 select array[1,3] <@ arrayrange(array[1,2], array[2,1]);
 
---
--- Check behavior when subtype lacks a hash function
---
-
-create type cashrange as range (subtype = money);
-
-set enable_sort = off;  -- try to make it pick a hash setop implementation
-
-select '(2,5)'::cashrange except select '(5,6)'::cashrange;
-
-reset enable_sort;
-
 --
 -- OUT/INOUT/TABLE functions
 --
